#!/usr/bin/env python

'''
wbutil/math/__init__.py

Top-level manifest of the math utilities.

Will Badart <wbadart@live.com
created: FEB 2018
'''

from . import stat
from . import naivebayes

from .naivebayes import *
from .stat import *
